/***************************************************************************
                          ToyLoad.c

    Contains the code for processing .lod files.
                             -------------------
    begin                : Sat Jan 25 2003
    copyright            : (C) 2003 by Tyler Montbriand
    student #            : 200200370
    class                : CS330
    email                : tsm@accesscomm.ca

   TODO - convert from from Explodee calls to LineToType calls
    
 ***************************************************************************/

#include<stdlib.h>
#include<stdio.h>
#include<string.h>
#include<ctype.h>

#include "ToySegment.h"
#include "ToySegmentList.h"
#include "ToyProcess.h"
#include "ToyMemType.h"
#include "ToyMem.h"
#include "ToyMem_Internal.h"
#include "toyparse.h"
#include "ToyLoad.h"

/*! \brief The magic number that designates a .lod file header */
const char  *DataType_Magic[]       = {  "8585",NULL           };
/*! \brief Two kinds of integers are recognized, raw and typed */
const char  *DataType_Integer[]     = {  "%u","#%u",NULL       };
/*! \brief Strings designated by a " */
const char  *DataType_String[]      = {  "\"%s",NULL           };
/*! \brief Table of valid instruction types */
const char  *DataType_Instruction[] = {  "RUN","HALT",NULL     };
/*! \brief Table of string tables of recognized command types.  Used by parser */
const char **DataType_List[]        = { DataType_Magic,
                                        DataType_Integer,
                                        DataType_String,
                                        DataType_Instruction,
                                        NULL                  };

 
/*! \brief  Retrieves the next line of tokens from a file being parsed. */
/**
 *  Returns NULL on failure.
 */
TokenType *GetTokenLine(FILE *fp, int *n);

/*! \brief  Attempts to process a .lod file */
/** Loads into the specifiedprocess in the specified memory address.
 *  \sa TMM_ParseFile()
 */
ErrMessage_t TMM_Load(ToyMM *tm, int process, const char *filename, int addr)
{
  int n           =-1;
  int numSegments =-1;
  int segSizes[4] ={-1,-1,-1,-1};
  FILE *fileIn    =NULL;
  TokenType *tarray=NULL;
  /*  Can't open nonexistent files,
      nor use nonexistent toy-mm's  */
  if((tm==NULL)||(filename==NULL))          return(ERR_UNKNOWN);
  else if((process<0)||(process>=MAX_PROCESSES))
  {
    return(ERR_UNKNOWN);
  }
  else if(tm->processList[process]!=NULL)
  {
    return(ERR_PROCESSEXISTS);
  }

  fileIn=fopen(filename,"r");

  if(fileIn==NULL)  return(ERR_NOFILE); /*  Couldn't open file*/

  /*  Get 1st line.  Should be magic #. */
  tarray=GetTokenLine(fileIn,&n);
  if(CheckLineSyntax(tarray,DATA_MAGIC,0,TYPE_ENDCLOSE)<0)
  {
    fclose(fileIn);    
    FreeType(tarray);
    return(ERR_INVALIDFILE);
  }
  FreeType(tarray);

  /*  Get 2nd line.  Should be # segments.  */
  tarray=GetTokenLine(fileIn,&n);
  if(CheckLineSyntax(tarray,DATA_INTEGER,0,TYPE_ENDCLOSE)<0)
  {
    fclose(fileIn);    
    FreeType(tarray);
    return(ERR_INVALIDFILE);
  }
  else if(sscanf(tarray[0].token,"%d",&numSegments)!=1)
  {
    fclose(fileIn);    
    FreeType(tarray);
    return(ERR_INVALIDFILE);
  }
  else if((numSegments<=0)||(numSegments>4))
  {
    fclose(fileIn);    
    FreeType(tarray);
    return(ERR_INVALIDFILE);
  }

  FreeType(tarray);

  /*  Get the segment sizes.  Should be like "1 3 8 4"  */
  tarray=GetTokenLine(fileIn,&n);
  if(tarray==NULL)  return(ERR_INVALIDFILE);
  else if(n!=numSegments)
  {
    fclose(fileIn);    
    FreeType(tarray);
    return(ERR_INVALIDINSTRUCTION);
  }
  
  for(n=0; n<numSegments; n++)
  {
    if(tarray[n].type!=DATA_INTEGER)
    {
      fclose(fileIn);      
      FreeType(tarray);
      return(ERR_INVALIDFILE);
    }
    else if(sscanf(tarray[n].token,"%d",&segSizes[n])!=1)
    {
      fclose(fileIn);      
      FreeType(tarray);
      return(ERR_INVALIDFILE);      
    }
    else if(segSizes[n]<=0)
    {
      fclose(fileIn);      
      FreeType(tarray);
      return(ERR_INVALIDFILE);      
    }
  }
  FreeType(tarray);

  /*  Allocate the process. */
  if(TMM_AllocateProcessAt(tm,process,addr,numSegments,segSizes)!=ERR_NONE)
  {
    fclose(fileIn);
    return(ERR_INSUFFICIENTMEM);
  }

  { /*  OK, now we're cookin'.  Load stuff into memory! */
    int curseg    =-1;
    int i         =-1;
    int numStrings=-1;      /*  # of strings in list  */
    int segpos    =-1;      /*  Position in segment   */
    int abspos    =-1;      /*  absolute memory position  */
    SegList *list = NULL;   /*  Current segment       */

    segpos=0;
    curseg=0;
    list=tm->processList[process]->segments;  /*Get the list of segments*/
    list=TMM_SeekSegListHead(list);

    /*  Get the next line of input  */
    tarray=GetTokenLine(fileIn,&numStrings);    
    while(tarray)
    {
      for(i=0; i<numStrings; i++)
      {
        /*  If we've reached the end of this segment,
            advance to the next one.                */
        if(segpos>=list->seg.len)
        {
          segpos=0;
          list=list->next;
          curseg++;
        }

        /*  If there is no next segment, we're out of memory. */
        if(list==NULL)
        {
          /*  Out of segment space! */
          TMM_TerminateProcess(tm,process);
          FreeType(tarray);
          return(ERR_DATAOVERFLOW);
        }

        /* Calculate our position in memory */
        abspos=segpos+list->seg.pos;

        switch(curseg)
        {
        case SEG_CODE:
          if(tarray[i].type!=DATA_INSTRUCTION)
          {
            FreeType(tarray);
            fclose(fileIn);
            TMM_TerminateProcess(tm,process);
            return(ERR_INVALIDINSTRUCTION);
          }
          break;

        case SEG_DATA:
          if((tarray[i].type!=DATA_INTEGER)&&(tarray[i].type!=DATA_STRING))
          {
            FreeType(tarray);
            fclose(fileIn);
            TMM_TerminateProcess(tm,process);
            return(ERR_INVALIDDATA);
          }
          break;

        default:
          FreeType(tarray);
          fclose(fileIn);
          TMM_TerminateProcess(tm,process);
          return(ERR_DATAOVERFLOW);
        }        

        if(tarray[i].type==DATA_ERROR)  /*  If the parser doesn't recognize it  */
        {
          FreeType(tarray);
          fclose(fileIn);
          return(ERR_INVALIDDATA);
        }

        tm->mem[abspos].type=(DataType_e)tarray[i].type;

        switch(tarray[i].type)
        {
        default:
          printf("Unknown type %d\n",tarray[i].type);
          printf("Email me at tsm@accesscomm.ca and tell me how you did this.\n");
          abort();
          break;
          
        case DATA_INTEGER:
          {
            int val;
            sscanf(tarray[i].token+1,"%d",&val);
            tm->mem[abspos].integer.value=val;
          }
          break;
        case DATA_INSTRUCTION:
          tm->mem[abspos].integer.value=tarray[i].subtype;
          break;

        case DATA_STRING:
          tm->mem[abspos].string.value[0]='\0';
          if(strlen(tarray[i].token)>MAX_STRLEN)
          {
            FreeType(tarray);
            fclose(fileIn);
            return(ERR_INVALIDDATA);
            break;
          }

          strcpy(tm->mem[abspos].string.value,tarray[i].token+1);
          break;
        }

        segpos++;
      }
      FreeType(tarray);  
      tarray=GetTokenLine(fileIn,&numStrings);
    }
  }
  
  
  fclose(fileIn);
  return(ERR_NONE);
}


/*! \brief Retrieves a line of text from a file, and turns it into an array of token-types.*/
/**
 *  Returns the array on success, NULL on failure.
 */
TokenType *GetTokenLine(FILE *fp, int *n)
{
  char buf[300];
  TokenType *tarray=NULL;
  buf[0]='\0';

  if(feof(fp))
  {
    if(n!=NULL) (*n)=0;
    return(NULL);
  }
  fgets(buf,300,fp);    /*  Get text                  */
  tarray=LineToTypes(buf,DataType_List);

  if((tarray!=NULL)&&(n!=NULL))
  {
    for((*n)=0; tarray[(*n)].token!=NULL; (*n)++);

    return(tarray);
  }
  else
  {
    (*n)=0;
    return(NULL);
  }
}
